/*-
 * Copyright (C)2021 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2021 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: qr.c,v 1.3 2021/08/07 10:17:13 babolo Exp $"

#define BLIN_COMPAT     3

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <regex.h>
#include <qrcodegen/qrcodegen.h>
#include <babolo/BLINflag.h>
#include "qrben.h"

#define QRBEN_MAXMODE (qrcodegen_Mode_KANJI + 1)
#define QSEGO(SEG) do {                                                                               \
    ifBLIN_QO4 {                                                                                      \
        for (u_int q_sego_inernal = 0; q_sego_inernal < (SEG)->spn; ++q_sego_inernal) {               \
            if  (qrcodegen_Mode_ECI == (SEG)->sp[q_sego_inernal].qlas) {                              \
                ifBLIN_QX4( "=#=%u s%u [%u]%s=%02X%02X%02X"                                           \
                          , q_sego_inernal                                                            \
                          , (SEG)->sp[q_sego_inernal].start                                           \
                          , (SEG)->sp[q_sego_inernal + 1].start - (SEG)->sp[q_sego_inernal].start     \
                          , qrben_qlnm[qrcodegen_Mode_ECI]                                            \
                          , text[(SEG)->sp[q_sego_inernal].start]                                     \
                          , text[(SEG)->sp[q_sego_inernal].start + 1]                                 \
                          , text[(SEG)->sp[q_sego_inernal].start + 2]                                 \
                          );                                                                          \
            } else {                                                                                  \
                ifBLIN_QX4( "=#=%u s%u [%u]%s=%.*s~"                                                  \
                          , q_sego_inernal                                                            \
                          , (SEG)->sp[q_sego_inernal].start                                           \
                          , (SEG)->sp[q_sego_inernal + 1].start - (SEG)->sp[q_sego_inernal].start     \
                          , (QRBEN_MAXMODE > (SEG)->sp[q_sego_inernal].qlas)                          \
                          ? qrben_qlnm[(SEG)->sp[q_sego_inernal].qlas]                                \
                          : qrben_qlnm[0]                                                             \
                          , (SEG)->sp[q_sego_inernal + 1].start - (SEG)->sp[q_sego_inernal].start     \
                          , &text[(SEG)->sp[q_sego_inernal].start]                                    \
                          );                                                                          \
        }   }                                                                                         \
        ifBLIN_QX4("=#=%u s%u", (SEG)->spn, (SEG)->sp[(SEG)->spn].start);                             \
    }                                                                                                 \
} while(0)

const char    qrben_num[]      = "0123456789";
const char    qrben_alpha[]    = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ $%*+-./:";
const char    qrben_qlnm[][4]  = {"nil", "NUM", "ALN", "non", "BYT", "non", "non", "ECI", "KAN"};
const uint8_t qrben_qlass[128] =
{  0, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
,  2, 4, 4, 4,  2, 2, 4, 4,  4, 4, 2, 2,  4, 2, 2, 2
,  1, 1, 1, 1,  1, 1, 1, 1,  1, 1, 2, 4,  4, 4, 4, 4

,  4, 2, 2, 2,  2, 2, 2, 2,  2, 2, 2, 2,  2, 2, 2, 2
,  2, 2, 2, 2,  2, 2, 2, 2,  2, 2, 2, 4,  4, 4, 4, 4
,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
};

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****     text                                                             ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_addseg(qrben_conf *cf, u_int16_t len, const char *text, qrben_mode mode) {                 /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    uint8_t *segdata;
    int      seglen;
    char    *copy;
    int      ex = EX_OK;

    ifBLIN_QX3("+ %08X m%d [%d] %d(%d)", cf->flags, mode, len, cf->nseg, cf->smax);
    if  (cf->nseg >= cf->smax) {
        cf->segs = realloc(cf->segs, (cf->smax = !cf->smax ? 4 : (2 * cf->smax)) * sizeof(qrben_seg));
        QRBEN_W(!cf->segs,                                  ("realloc segs x %d", cf->smax), EX_OSERR);
    }
    seglen = qrcodegen_calcSegmentBufferSize(mode, (qrcodegen_Mode_ECI == mode) ? 0 : len);
    QRBEN_X((4096 < seglen),               ("calcSegmentBufferSize %d", seglen), EDOOFUS, EX_SOFTWARE);
    QRBEN_W(!(segdata = malloc(seglen)),                            ("malloc x %d", seglen), EX_OSERR);
    switch(mode) {
    case qrcodegen_Mode_NUMERIC:
        QRBEN_W(!(copy = strndup(text, len)),                           ("strndup %d", len), EX_OSERR);
        QRBEN_X( !qrcodegen_isNumeric(copy)
               , ("%d is not numeric, text=%s~", cf->nseg, copy)
               , EDOOFUS
               , EX_USAGE
               );
        cf->segs[cf->nseg] = qrcodegen_makeNumeric(copy, segdata);
        free(copy);
        break;
    case qrcodegen_Mode_ALPHANUMERIC:
        QRBEN_W(!(copy = strndup(text, len)),                           ("strndup %d", len), EX_OSERR);
        QRBEN_X( !qrcodegen_isAlphanumeric(copy)
               , ("%d is not alfanumeric, text=%s~", cf->nseg, copy)
               , EDOOFUS
               , EX_USAGE
               );
        cf->segs[cf->nseg] = qrcodegen_makeAlphanumeric(copy, segdata);
        free(copy);
        break;
    case qrcodegen_Mode_BYTE:
        cf->segs[cf->nseg] = qrcodegen_makeBytes((const uint8_t *)text, len, segdata);
        break;
    case qrcodegen_Mode_ECI:
        ifBLIN_QX0("For ECI use qrben_addeci() instead qrben_addseg()");
        ex = EX_USAGE;
        errno = EDOM;
        goto out;
    default:
        ifBLIN_QX0("mode %d unsuppoted", mode);
        ex = EX_USAGE;
        errno = EDOM;
        goto out;
    }
    ifBLIN_QX4( "%s[%d]=>%db"
              , (QRBEN_MAXMODE > cf->segs[cf->nseg].mode)
              ? qrben_qlnm[cf->segs[cf->nseg].mode]
              : qrben_qlnm[0]
              , cf->segs[cf->nseg].numChars
              , cf->segs[cf->nseg].bitLength
              );
    ifBLIN_QO4 blin_dumb(4, cf->segs[cf->nseg].data, seglen, "");
    cf->nseg++;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    ECI num                                                                 ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_addeci(qrben_conf *cf, u_int num) {                                                        /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    uint8_t *segdata;
    int      seglen;
    int      ex = EX_OK;

    ifBLIN_QX3("+ %08X %u %d(%d)", cf->flags, num, cf->nseg, cf->smax);
    if  (cf->nseg >= cf->smax) {
        cf->segs = realloc(cf->segs, (cf->smax = !cf->smax ? 4 : (2 * cf->smax)) * sizeof(qrben_seg));
        QRBEN_W(!cf->segs,                                  ("realloc segs x %d", cf->smax), EX_OSERR);
    }
    seglen = qrcodegen_calcSegmentBufferSize(qrcodegen_Mode_ECI, 0);
    QRBEN_X((4096 < seglen),               ("calcSegmentBufferSize %d", seglen), EDOOFUS, EX_SOFTWARE);
    QRBEN_W(!(segdata = malloc(seglen)),                            ("malloc x %d", seglen), EX_OSERR);
    cf->segs[cf->nseg] = qrcodegen_makeEci(num, segdata);
    ifBLIN_QX4( "%s[%d]=>%db"
              , (QRBEN_MAXMODE > cf->segs[cf->nseg].mode)
              ? qrben_qlnm[cf->segs[cf->nseg].mode]
              : qrben_qlnm[0]
              , cf->segs[cf->nseg].numChars
              , cf->segs[cf->nseg].bitLength
              );
    ifBLIN_QO4 blin_dumb(4, cf->segs[cf->nseg].data, seglen, "");
    cf->nseg++;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static u_int
/*****************************************************************************************************
                                                                                                  ****/
testq(qrben_conf *cf, char c) {                                                                  /****
                                                                                                  ****
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    u_int e;

    (void)cf;
    if  ((0 > c) || (128 <= c)) {
        e = qrcodegen_Mode_BYTE;
    } else {
        e = qrben_qlass[c & 0x0FF];
    }
    return(e);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****                                                                 ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_segsplen(qrben_conf *cf, qrben_segsp *segs, u_int versize) {                               /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    int         ex = EX_OK;
    u_int       nm[3] = {0, 4, 7};
    u_int       am[2] = {0, 6};
    u_int       seglen;
    u_int       modulo;

    ifBLIN_QX3("+ %08X %d(%d)", cf->flags, segs->spn, versize);
    for (u_int i = 0; i < segs->spn; ++i) {
        seglen = segs->sp[i + 1].start - segs->sp[i].start;
        switch(segs->sp[i].qlas) {
        case qrcodegen_Mode_NUMERIC:
            modulo = seglen % 3;
            seglen /= 3;
            ex += (seglen * 10) + nm[modulo] + 14;
            if  (9 < versize) ex += 2;
            if  (26 < versize) ex += 2;
            ifBLIN_QX4("%u NUM %u+%u=>%d", i, seglen, modulo, ex);
            break;
        case qrcodegen_Mode_ALPHANUMERIC:
            modulo = seglen % 2;
            seglen /= 2;
            ex += (seglen * 11) + am[modulo] + 13;
            if  (9 < versize) ex += 2;
            if  (26 < versize) ex += 2;
            ifBLIN_QX4("%u ALN %u+%u=>%d", i, seglen, modulo, ex);
            break;
        case qrcodegen_Mode_BYTE:
            ex += (seglen * 8) + 12;
            if  (9 < versize) ex += 8;
            ifBLIN_QX4("%u BYT %u=>%d", i, seglen, ex);
            break;
        case qrcodegen_Mode_ECI:
            ifBLIN_QX0("mode ECI in plain text unsupported");
            ex = -EX_USAGE;
            errno = EINVAL;
            goto out;
        default:
            ifBLIN_QX0("mode %d unsuppoted", segs->sp[i].qlas);
            ex = -EX_USAGE;
            errno = EDOM;
            goto out;
    }   }
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****                                                                    ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_aseg(qrben_conf *cf, u_int16_t len, const char *text, u_int versize) {                     /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    int          seg0len = 0;
    int          segtlen = 0;
    int          seg1len = 0;
    u_int        qlasnow;
    u_int        ntrans;
    qrben_segsp *seg0    = NULL;
    qrben_segsp *segt    = NULL;
    qrben_segsp *seg1    = NULL;
    int          ex      = EX_OK;

    ifBLIN_QX3("+ %08X [%d] %d(%d)", cf->flags, len, cf->nseg, cf->smax);
    ntrans = 0;
    qlasnow = 0;
    for (u_int i = 0; i < len; ++i) {                  /*     */
        if  (qlasnow != testq(cf, text[i])) ++ntrans;
        qlasnow = testq(cf, text[i]);
    }
    seg0 = malloc((1 + ntrans) * sizeof(qrben_segg) + sizeof(qrben_segsp));
    QRBEN_W(!seg0,                                                     ("segsp %u", ntrans), EX_OSERR);
    seg0->spn = ntrans;
    seg0->sp[ntrans].start = len;
    qlasnow = 0;
    for (u_int i = 0, n = 0; i < len; ++i) {                     /*      */
        if  (qlasnow != testq(cf, text[i])) {
            seg0->sp[n].start = i;
            seg0->sp[n].qlas = qlasnow = testq(cf, text[i]);
            ifBLIN_QX4( "%u s%u %s(%u)"
                      , n
                      , seg0->sp[n].start
                      , (QRBEN_MAXMODE > seg0->sp[n].qlas)
                      ? qrben_qlnm[seg0->sp[n].qlas]
                      : qrben_qlnm[0]
                      , seg0->sp[n].qlas
                      );
            ++n;
    }   }
    QSEGO(seg0);
    do { /*  ,    seg0  segt,  ,   seg0 */
        QRBEN_W((0 > (seg0len = qrben_segsplen(cf, seg0, versize))),               ("segsplen"), -seg0len);
        QSEGO(seg0);
        for (u_int i = 1; i < seg0->spn; ++i) {    /*        */
            segt = malloc(seg0->spn * sizeof(qrben_segg) + sizeof(qrben_segsp));
            QRBEN_W(!segt,                                          ("segsp %u", seg0->spn), EX_OSERR);
            segt->spn = 0;
            segt->sp[segt->spn].start = len;
            for (u_int j = 0, k = 0; j < seg0->spn; ++j) {  /*     */
                if  (j == (i - 1)) {
                    segt->sp[k].start = seg0->sp[j].start;
                    segt->sp[k].qlas = seg0->sp[j].qlas;
                    if  (seg0->sp[i].qlas > segt->sp[k].qlas) segt->sp[k].qlas = seg0->sp[i].qlas;
                    if  (!k || (segt->sp[k].qlas != segt->sp[k - 1].qlas)) {
                        ifBLIN_QX4("%u: s%u q%u", k, segt->sp[k].start, segt->sp[k].qlas);
                        segt->spn = ++k;
                    }
                    segt->sp[segt->spn].start = len;
                } else if ((j != i) && (!k || (segt->sp[k - 1].qlas != seg0->sp[j].qlas))) {
                    segt->sp[k].start = seg0->sp[j].start;
                    segt->sp[k].qlas = seg0->sp[j].qlas;
                    ifBLIN_QX4("%u: s%u q%u", k, segt->sp[k].start, segt->sp[k].qlas);
                    segt->spn = ++k;
                    segt->sp[segt->spn].start = len;
            }   }
            QSEGO(segt);
            QRBEN_W((0 > (segtlen = qrben_segsplen(cf, segt, versize))),       ("segsplen"), -segtlen);
            ifBLIN_QX4("0=%d %"BLIN_X"=%d =%d", seg0len, BLIN_I(seg1), seg1len, segtlen);
            if  ((segtlen < seg0len) && (!seg1 || (segtlen < seg1len))) {
                free(seg1);
                seg1 = segt;
                segt = NULL;
                seg1len = segtlen;
                segtlen = 0;
            } else {
                free(segt);
                segt = NULL;
                segtlen = 0;
        }   }
        if  (!!seg1 && (seg1len < seg0len)) {
            free(seg0);
            seg0 = seg1;
            seg1 = NULL;
            seg0len = seg1len;
            seg1len = 0;
        } else {
            seg1 = seg0;
            seg0 = NULL;
            break;
        }
    } while(1);
    QSEGO(seg1);
    for (u_int n = 0; n < seg1->spn; ++n) {
        qrben_addseg( cf
                    , seg1->sp[n + 1].start - seg1->sp[n].start
                    , &text[seg1->sp[n].start]
                    , seg1->sp[n].qlas
                    );
    }
    free(seg1);
    seg1 = NULL;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

int
/*****************************************************************************************************
 *****************************************************************************************************
 ****    , ,          ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_pseg(qrben_conf *cf, u_int16_t len, const char *text, u_int versize) {                     /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    char         erbuf[256];
    regmatch_t   remat;
    regex_t      reco;
    char        *copy;
    const char  *rex = "^s?(f|ht)tps?://[-0-9a-zA-Z.]+";
    int          ex  = EX_OK;

    ifBLIN_QX3("+ %08X [%d] %d(%d)", cf->flags, len, cf->nseg, cf->smax);
    ex = regcomp(&reco, rex, REG_EXTENDED | REG_ICASE);
    if  (!!ex) {
        regerror(ex, &reco, erbuf, 256);
        ifBLIN_QX0("regcomp: %s", erbuf);
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    }
    ex = regexec(&reco, text, 1, &remat, 0);
    if  (REG_NOMATCH == ex) {
        ex = EX_OK;
        QRBEN_O(!!(ex = qrben_aseg(cf, len, text, versize)),                                 ("aseg"));
    } else if (!!ex) {
        regerror(ex, &reco, erbuf, 256);
        ifBLIN_QX0("regcomp: %s", erbuf);
        errno = EDOOFUS;
        ex = EX_SOFTWARE;
        goto out;
    } else {
        QRBEN_W(!(copy = strndup(text, len)),                           ("strndup %d", len), EX_OSERR);
        for (u_int i = remat.rm_so; i < remat.rm_eo; ++i) {
            if  (('a' <= copy[i]) && ('z' >= copy[i])) copy[i] &= 0x0DF;
        }
        QRBEN_O(!!(ex = qrben_aseg(cf, len, copy, versize)),                                 ("aseg"));
        free(copy);
    }
    regfree(&reco);
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

qrben_code *
/*****************************************************************************************************
 *****************************************************************************************************
 ****   QR                                                                        ****
 *****************************************************************************************************
                                                                                                  ****/
qrben_gen(qrben_conf *cf) {                                                                      /****
                                                                                                  ****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define blin_internal_flags (cf->flags)
    qrben_code *qrcode = NULL;
    uint8_t    *buf    = NULL;
    uint8_t    *tmp    = NULL;
    int         msz    = qrcodegen_BUFFER_LEN_FOR_VERSION(cf->versize);
    int         ex;

    QRBEN_W(!(buf = malloc(msz)),                               ("buf ver %d", cf->versize), EX_OSERR);
    QRBEN_W(!(tmp = malloc(msz)),                               ("tmp ver %d", cf->versize), EX_OSERR);
    ex = !qrcodegen_encodeSegmentsAdvanced( cf->segs
                                          , cf->nseg
                                          , cf->errc
                                          ,   (QRBEN_FIXSIZE & cf->flags)
                                            ? cf->versize
                                            : qrcodegen_VERSION_MIN
                                          , cf->versize
                                          , qrcodegen_Mask_AUTO
                                          , !(QRBEN_FIXERRC & cf->flags)
                                          , tmp
                                          , buf
                                          );
    QRBEN_X(!!ex,                          ("qrcodegen_encodeSegmentsAdvanced"), EDOOFUS, EX_SOFTWARE);
    free(tmp);
    tmp = NULL;
    qrcode = malloc(qrcodegen_getSize(buf) * qrcodegen_getSize(buf) + sizeof(qrben_code));
    QRBEN_W(!qrcode,                                            ("versize %d", cf->versize), EX_OSERR);
    qrcode->size = qrcodegen_getSize(buf);
    for (int y = 0; y < qrcode->size; y++) {
        for (int x = 0; x < qrcode->size; x++) {
            qrcode->code[qrcode->size * y + x] = qrcodegen_getModule(buf, x, y) ? '1' : '0';
    }   }
out:
    free(buf);
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(qrcode));
    return(qrcode);
#   undef blin_internal_flags
}
